searchData={"content_type":"text/markdown","items":[{"doc":"A module that provides a interface to Erlang's time functions.\n\nIn production, all functions are proxied to the `:erlang` module directly.\n\nIn tests, the functions are proxied to `Klotho.Mock`, and the time \"flow\"\ncan be controlled by calling `Klotho.Mock` functions.","ref":"Klotho.html","title":"Klotho","type":"module"},{"doc":"","ref":"Klotho.html#cancel_timer/1","title":"Klotho.cancel_timer/1","type":"function"},{"doc":"","ref":"Klotho.html#cancel_timer/2","title":"Klotho.cancel_timer/2","type":"function"},{"doc":"","ref":"Klotho.html#monotonic_time/0","title":"Klotho.monotonic_time/0","type":"function"},{"doc":"","ref":"Klotho.html#monotonic_time/1","title":"Klotho.monotonic_time/1","type":"function"},{"doc":"","ref":"Klotho.html#read_timer/1","title":"Klotho.read_timer/1","type":"function"},{"doc":"","ref":"Klotho.html#send_after/3","title":"Klotho.send_after/3","type":"function"},{"doc":"","ref":"Klotho.html#send_after/4","title":"Klotho.send_after/4","type":"function"},{"doc":"","ref":"Klotho.html#set_backend/1","title":"Klotho.set_backend/1","type":"function"},{"doc":"","ref":"Klotho.html#start_timer/3","title":"Klotho.start_timer/3","type":"function"},{"doc":"","ref":"Klotho.html#start_timer/4","title":"Klotho.start_timer/4","type":"function"},{"doc":"","ref":"Klotho.html#system_time/0","title":"Klotho.system_time/0","type":"function"},{"doc":"","ref":"Klotho.html#system_time/1","title":"Klotho.system_time/1","type":"function"},{"doc":"","ref":"Klotho.html#time_offset/0","title":"Klotho.time_offset/0","type":"function"},{"doc":"","ref":"Klotho.html#time_offset/1","title":"Klotho.time_offset/1","type":"function"},{"doc":"This is the backend is used in `:test` environment.\n\nThe module gives control of the mocked time and provides inspection\nof triggered timers.","ref":"Klotho.Mock.html","title":"Klotho.Mock","type":"module"},{"doc":"Freeze time. In `frozen` mode, all calls to `monotonic_time` and `system_time`\nwill return the same value. Timers will be not triggered unless warp is called.\nThe function is idempotent.","ref":"Klotho.Mock.html#freeze/0","title":"Klotho.Mock.freeze/0","type":"function"},{"doc":"Reset the state of the time server. This\n* resets the time to the actual current time;\n* cleans timer history;\n* resets the mode to `running`;\n* cancels all timers.","ref":"Klotho.Mock.html#reset/0","title":"Klotho.Mock.reset/0","type":"function"},{"doc":"Return history of all messages (`%Klotho.Mock.TimerMsg{}`) sent by triggered timers.\nMost recent messages are first.","ref":"Klotho.Mock.html#timer_event_history/0","title":"Klotho.Mock.timer_event_history/0","type":"function"},{"doc":"Unreeze time. In `running` mode, all calls to `monotonic_time` and `system_time` will\nstart produce values increasing with \"normal\" monotonic time pace.\nTimers will be triggered according to their schedule.","ref":"Klotho.Mock.html#unfreeze/0","title":"Klotho.Mock.unfreeze/0","type":"function"},{"doc":"Warp time forward by `timer_interval` milliseconds.\nRegardless of the current mode (frozen or running), all the timers\nthat are due to trigger within the next `timer_interval` milliseconds\nwill be triggered after warp.","ref":"Klotho.Mock.html#warp_by/1","title":"Klotho.Mock.warp_by/1","type":"function"},{"doc":"Warp time forward by `timer_interval` in `unit`.\nRegardless of the current mode (frozen or running), all the timers\nthat are due to trigger within the next `timer_interval` milliseconds\nwill be triggered after warp.","ref":"Klotho.Mock.html#warp_by/2","title":"Klotho.Mock.warp_by/2","type":"function"},{"doc":"Struct to represent a timer message.","ref":"Klotho.Mock.TimerMsg.html","title":"Klotho.Mock.TimerMsg","type":"module"},{"doc":"","ref":"Klotho.Mock.TimerMsg.html#t:t/0","title":"Klotho.Mock.TimerMsg.t/0","type":"type"},{"doc":"# `Klotho` usage","ref":"usage.html","title":"Klotho usage","type":"extras"},{"doc":"Testing code that deals with time, timeouts, and timers is hard. The main reason is that\ngetting time and setting timers is often not treated as a public contract, but it actually is.\n\nOften, to test such code, one provides custom significantly reduced timeouts and uses `:timer.sleep/1`.\nThis approach has several drawbacks:\n* it makes the tests slow\n* if we use lagre timeouts, it makes the tests even slower\n* if we use small timeouts, it makes the tests flaky\n\nOne of the approaches to testing without sleeps is injecting time-related functions directly,\nthus making the contract explicitly public. However, this makes the code much more\ncomplex and harder to read.\n\n`Klotho` takes a different approach. It injects timer-based functions globally.\nWith `Klotho` you do not use time-related functions directly, but instead, you use `Klotho` functions that wrap the original ones in production code.\n\nIn tests, these functions are replaced with a mock implementation that allows controlling time \"flow\".\nSee `Klotho.Mock` for details.","ref":"usage.html#general-cosiderations","title":"General cosiderations - Klotho usage","type":"extras"},{"doc":"Assume we have a module implementing a simple \"timeout map\" functionality. It allows setting\na timeout for a key and records are automatically removed from the map when the timeout expires.\n\n\nA possible (and a bit naive) implementation could look like this:\n\n```elixir\ndefmodule TimeoutCache do\n  @moduledoc false\n\n  use GenServer\n\n  # API\n\n  def start_link() do\n    GenServer.start_link(__MODULE__, [])\n  end\n\n  def set(pid, key, value, timeout) do\n    GenServer.call(pid, {:set, key, value, timeout})\n  end\n\n  def get(pid, key) do\n    GenServer.call(pid, {:get, key})\n  end\n\n  # gen_server\n\n  def init([]) do\n    {:ok, %{}}\n  end\n\n  def handle_call({:set, key, value, timeout}, _from, state) do\n    new_st =\n      state\n      |> maybe_delete(key)\n      |> put_new(key, value, timeout)\n\n    {:reply, :ok, new_st}\n  end\n\n  def handle_call({:get, key}, _from, state) do\n    {:reply, get_value(state, key), state}\n  end\n\n  def handle_info({:timeout, ref, key}, state) do\n    new_st = maybe_delete_timeout(state, key, ref)\n    {:noreply, new_st}\n  end\n\n  # private\n\n  defp maybe_delete(state, key) do\n    case state do\n      %{^key => {_value, ref}} ->\n        :erlang.cancel_timer(ref)\n        Map.delete(state, key)\n\n      _ ->\n        state\n    end\n  end\n\n  defp put_new(state, key, value, timeout) do\n    ref = :erlang.start_timer(timeout, self(), key)\n    Map.put(state, key, {value, ref})\n  end\n\n  defp maybe_delete_timeout(state, key, ref) do\n    case state do\n      %{^key => {_value, ^ref}} ->\n        Map.delete(state, key)\n\n      _ ->\n        state\n    end\n  end\n\n  defp get_value(state, key) do\n    case state do\n      %{^key => {value, _ref}} ->\n        {:ok, value}\n\n      _ ->\n        :not_found\n    end\n  end\nend\n```\n\nHow do we test that keys are actually removed from the map after the timeout expires?\nA possible test suite could look like this:\n\n```elixir\ndefmodule TimeoutCacheTest do\n  use ExUnit.Case\n\n  setup do\n    {:ok, pid} = TimeoutCache.start_link()\n    {:ok, pid: pid}\n  end\n\n  test \"set and get\", %{pid: pid} do\n    TimeoutCache.set(pid, :foo, :bar, 100)\n    assert {:ok, :bar} == TimeoutCache.get(pid, :foo)\n  end\n\n  test \"get after timeout\", %{pid: pid} do\n    TimeoutCache.set(pid, :foo, :bar, 100)\n    :timer.sleep(200)\n    assert :not_found == TimeoutCache.get(pid, :foo)\n  end\n\n  test \"renew lifetime\", %{pid: pid} do\n    TimeoutCache.set(pid, :foo, :bar1, 100)\n    :timer.sleep(50)\n    TimeoutCache.set(pid, :foo, :bar2, 100)\n    :timer.sleep(70)\n    assert {:ok, :bar2} == TimeoutCache.get(pid, :foo)\n  end\nend\n```\n\nThis test suite is slow and flaky. It is slow because of `:timer.sleep/1` calls.\nAlso, if the test machine is under heavy load, the timeouts may expire later than expected, thus making the tests flaky.\nOn the other hand, if we increase the timeouts, the tests will become even slower.\n\nWith `Klotho` we may rewrite the implementation as follows:\n\n```elixir\ndefmodule TimeoutCache do\n  ...\n\n  defp maybe_delete(state, key) do\n    case state do\n      %{^key => {_value, ref}} ->\n        Klotho.cancel_timer(ref)\n        Map.delete(state, key)\n\n      _ ->\n        state\n    end\n  end\n\n  defp put_new(state, key, value, timeout) do\n    ref = Klotho.start_timer(timeout, self(), key)\n    Map.put(state, key, {value, ref})\n  end\n\n  ...\nend\n```\n\nWe just replaced `:erlang.cancel_timer/1` with `Klotho.cancel_timer/1` and `:erlang.start_timer/3`\nwith `Klotho.start_timer/3`. See [`timeout_cache.erl`](./test/support/timeout_cache.ex).\nNow we can rewrite the test suite as follows:\n\n```elixir\ndefmodule TimeoutCacheTest do\n  use ExUnit.Case\n\n  setup do\n    {:ok, pid} = TimeoutCache.start_link()\n    Klotho.Mock.reset()\n    Klotho.Mock.freeze()\n    {:ok, pid: pid}\n  end\n\n  test \"set and get\", %{pid: pid} do\n    TimeoutCache.set(pid, :foo, :bar, 1000)\n    assert {:ok, :bar} == TimeoutCache.get(pid, :foo)\n  end\n\n  test \"get after timeout\", %{pid: pid} do\n    TimeoutCache.set(pid, :foo, :bar, 1000)","ref":"usage.html#example","title":"Example - Klotho usage","type":"extras"},{"doc":"Klotho.Mock.warp_by(2000)\n    assert :not_found == TimeoutCache.get(pid, :foo)\n  end\n\n  test \"renew lifetime\", %{pid: pid} do\n    TimeoutCache.set(pid, :foo, :bar1, 1000)\n    Klotho.Mock.warp_by(500)\n    TimeoutCache.set(pid, :foo, :bar2, 1000)\n    Klotho.Mock.warp_by(700)\n    assert {:ok, :bar2} == TimeoutCache.get(pid, :foo)\n  end\nend\n```\n\nWe may use arbitrary timeouts in tests, just warping the time by the required amount. The code does not\nuse sleeps and runs fast. Also, the tests are not flaky anymore.\n\nSee the difference in the test execution time:\n\n```\n$ mix test test/klotho_timeout_cache_test.exs --trace\n\nKlotho.TimeoutCacheTest.Klotho [test/klotho_timeout_cache_test.exs]\n  * test set and get (1.7ms) [L#44]\n  * test renew lifetime (0.05ms) [L#55]\n  * test get after timeout (0.07ms) [L#49]\n\nKlotho.TimeoutCacheTest.Timer [test/klotho_timeout_cache_test.exs]\n  * test set and get (0.05ms) [L#12]\n  * test renew lifetime (121.8ms) [L#23]\n  * test get after timeout (200.9ms) [L#17]\n\nFinished in 0.3 seconds (0.00s async, 0.3s sync)\n6 tests, 0 failures\n```","ref":"usage.html#timers-whose-time-has-passed-are-triggered-in-the-end-of-warp","title":"Timers whose time has passed are triggered in the end of warp - Klotho usage","type":"extras"},{"doc":"`Klotho` does not intend to be suitable for any case and provide beam-wide time injection. It is\ndesigned to be used in a wide but still limited scope of cases when the code deals with some\nmedium-intensive self-contained time-related logic.\n\n* See `Klotho` for the full list of supported functions.\n* The library may not be suitable for testing logic with some high-frequency events or events with\nan order of millisecond latency because the time is managed by a `GenServer` process.\n* The library may not work well if code actively uses other modules with their own and\nintensive time-related logic.","ref":"usage.html#limitations","title":"Limitations - Klotho usage","type":"extras"},{"doc":"MIT License\n\nCopyright 2017-2023 Ilya Averyanov\n\nPermission is hereby granted, free of charge, to any person obtaining a copy\nof this software and associated documentation files (the \"Software\"), to deal\nin the Software without restriction, including without limitation the rights\nto use, copy, modify, merge, publish, distribute, sublicense, and/or sell\ncopies of the Software, and to permit persons to whom the Software is\nfurnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in all\ncopies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\nIMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\nFITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\nAUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\nLIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\nOUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE\nSOFTWARE.","ref":"license.html","title":"LICENSE","type":"extras"}]}